import std.array;
import std.conv;
import std.file;
import std.stdio;
import std.string;
import std.utf;

import filerange: FileRange;
import parsenumber : parseNumber;
import status : Severity, Status, StatusList, StatusMessage;
import tokenize : Token, Tokenizer;

// Enumeration of recognized instructions
enum Instruction: uint {
    Invalid = 0,
    _exit,
    _void,
    _goto,
    _transfer,
    _again,
    _return,
    _return_with,
    _swap,
    _push,
    _pop,
    _discard,
    _peek,
    _assign,
    _inc,
    _dec,
    _shl,
    _shr,
    _sar,
    _negate,
    _abs,
    _not,
    _add,
    _sub,
    _mul,
    _div,
    _mod,
    _rem,
    _and,
    _or,
    _xor,
    _clear_error,
    _write_char,
    _write_int,
    _read_char,
    _read_int,
}

// Enumeration of recognized condition arguments
enum ConditionArgument: uint {
    None = 0,
    _always,
    _if_zero,
    _if_nonzero,
    _if_positive,
    _if_not_positive,
    _if_negative,
    _if_not_negative,
    _if_carry,
    _if_not_carry,
    _if_divz,
    _if_not_divz,
    _if_wrapped,
    _if_not_wrapped,
    _if_error,
    _if_no_error,
}

// Enumeration of recognized node arguments
enum NodeArgument: uint {
    None = 0,
    _self,
    _root,
    _parent,
    _left,
    _right,
    _sibling,
    _leftmost,
    _rightmost,
    _next,
    _prev,
    _origin,
}

// What kind of value argument does a statement have?
enum ValueArgument: uint {
    None = 0, // No value argument
    Constant, // Constant value
    _top, // Use top of stack
    _acc, // Use accumulator
    _carry, // Use carry flag
    _overflow, // Use multiplication overflow
    _divz, // Use division by zero flag
    _wrapped, // Stack pointer overflow flag
    _error, // Use error flag
}

// Enumeration of named constants
// Does not include "no_error" because zero is used for the None member
enum NamedConstant: int {
    None = 0,
    _min = int.min,
    _max = int.max,
    _stack_size = Node.StackSize,
    _read_char_error = 0x01,
    _read_int_error = 0x02,
}

// Enumeration of possible ways for a statement to accept value arguments
enum ValueCount: uint {
    // Does not accept a value argument
    None = 0,
    // Accepts exactly one value argument
    One,
    // Accept one or more value arguments:
    // Expand `push "Yes";`
    // to `push "s"; push "e"; push "Y";`
    MultiForward,
    // Accept one or more value arguments:
    // Expand `write_char "Yes";`
    // to `write_char "Y"; write_char "e"; write_char "s";`
    MultiReverse,
}

// Indicates which instructions require or accept which arguments
struct InstructionInfo {
    enum NodeArgument: uint {No = 0, Yes = 1}
    enum ConditionArgument: uint {No = 0, Yes = 1}
    
    Instruction instruction;
    ValueCount valueCount; // Accepts a value argument?
    NodeArgument nodeArgument; // Optionally accepts a node getNode
    ConditionArgument conditionArgument; // Optionally accepts a jump condition
    
    // Nice syntax candy constructor
    this(T...)(in Instruction instruction, T attrs) {
        this.instruction = instruction;
        foreach(attr; attrs) {
            foreach(ref field; this.tupleof) {
                static if(is(typeof(attr) == typeof(field))) {
                    field = attr;
                }
            }
        }
    }
    
    // Get the InstructionInfo object given an Instruction.
    static auto get(in Instruction instruction) {
        const index = cast(size_t) instruction;
        assert(index < InstructionInfo.list.length);
        auto info = InstructionInfo.list[index];
        assert(info.instruction is instruction);
        return info;
    }
    
    // True when the instruction requires a value argument.
    @property bool needsValue() const {
        return this.valueCount !is ValueCount.None;
    }
    
    // True when a statement with this instruction accepts more than one
    // value argument.
    @property bool multiValue() const {
        return (
            this.valueCount is ValueCount.MultiForward ||
            this.valueCount is ValueCount.MultiReverse
        );
    }
    
    string toString() const {
        string result = this.instruction.to!string[1 .. $];
        if(this.nodeArgument) result ~= " <target>";
        if(this.valueCount is ValueCount.One) result ~= " <val>";
        else if(this.valueCount) result ~= " <val...>";
        if(this.conditionArgument) result ~= " <condition>";
        return result ~ ";";
    }
    
    // List associating Instruction enum members with InstructionInfo objects.
    static enum list = [
        InstructionInfo(Instruction.Invalid),
        InstructionInfo(Instruction._exit),
        InstructionInfo(Instruction._void),
        InstructionInfo(Instruction._goto, NodeArgument.Yes, ConditionArgument.Yes),
        InstructionInfo(Instruction._transfer, NodeArgument.Yes, ConditionArgument.Yes, ValueCount.One),
        InstructionInfo(Instruction._again, ConditionArgument.Yes),
        InstructionInfo(Instruction._return, ConditionArgument.Yes),
        InstructionInfo(Instruction._return_with, ConditionArgument.Yes, ValueCount.One),
        InstructionInfo(Instruction._swap, NodeArgument.Yes),
        InstructionInfo(Instruction._push, NodeArgument.Yes, ValueCount.MultiReverse),
        InstructionInfo(Instruction._pop, NodeArgument.Yes),
        InstructionInfo(Instruction._discard, NodeArgument.Yes),
        InstructionInfo(Instruction._peek, NodeArgument.Yes),
        InstructionInfo(Instruction._assign, NodeArgument.Yes, ValueCount.One),
        InstructionInfo(Instruction._inc),
        InstructionInfo(Instruction._dec),
        InstructionInfo(Instruction._shl, ValueCount.One),
        InstructionInfo(Instruction._shr, ValueCount.One),
        InstructionInfo(Instruction._sar, ValueCount.One),
        InstructionInfo(Instruction._negate),
        InstructionInfo(Instruction._abs),
        InstructionInfo(Instruction._not),
        InstructionInfo(Instruction._add, ValueCount.One),
        InstructionInfo(Instruction._sub, ValueCount.One),
        InstructionInfo(Instruction._mul, ValueCount.One),
        InstructionInfo(Instruction._div, ValueCount.One),
        InstructionInfo(Instruction._mod, ValueCount.One),
        InstructionInfo(Instruction._rem, ValueCount.One),
        InstructionInfo(Instruction._and, ValueCount.One),
        InstructionInfo(Instruction._or, ValueCount.One),
        InstructionInfo(Instruction._xor, ValueCount.One),
        InstructionInfo(Instruction._clear_error),
        InstructionInfo(Instruction._write_char, ValueCount.MultiForward),
        InstructionInfo(Instruction._write_int, ValueCount.One),
        InstructionInfo(Instruction._read_char),
        InstructionInfo(Instruction._read_int),
    ];
}

// Helper to implement getInstruction, getConditionArgument, and getNodeArgument
T getEnumMember(T)(in string name, in T fallback) {
    foreach(member; __traits(allMembers, T)) {
        if(member.length && member[0] == '_') {
            mixin(`T instruction = T.` ~ member ~ `;`);
            if(name == member[1 .. $]) return instruction;
        }
    }
    return fallback;
}

// Get the Instruction corresponding to a given name in source code
Instruction getInstruction(in string name) {
    return getEnumMember!Instruction(name, Instruction.Invalid);
}

// Get the ConditionArgument corresponding to a given name in source code
ConditionArgument getConditionArgument(in string name) {
    return getEnumMember!ConditionArgument(name, ConditionArgument.None);
}

// Get the NodeArgument corresponding to a given name in source code
NodeArgument getNodeArgument(in string name) {
    return getEnumMember!NodeArgument(name, NodeArgument.None);
}

// Get the ValueArgument corresponding to a given name in source code
ValueArgument getValueArgument(in string name) {
    return getEnumMember!ValueArgument(name, ValueArgument.None);
}

// Get the value of a named constant, corresponding to a given name in source code
// Returns zero when no named constant was recognized
// Note that the "no_error" constant (zero) is not recognized in this way
int getNamedConstant(in string name) {
    return cast(int) getEnumMember!NamedConstant(name, NamedConstant.None);
}

// Represents a single statement within a Jungle program
struct Statement {
    // Instruction associated with this statement
    Instruction instruction;
    // What node argument was provided in this statement, if any
    NodeArgument nodeArgument;
    // What condition argument was provided in this statement, if any
    ConditionArgument conditionArgument;
    // What type of value argument was provided, if any
    ValueArgument valueArgument;
    // In the case of a constant as a value argument, store the value here
    int constant;
    
    // Pointer to affected node (based on nodeArgument)
    Node* targetNode;
    
    // Location in source file (line number)
    size_t lineNumber;
    // Location in source file (column number)
    size_t colNumber;
    
    this(in Instruction instruction,
        in ValueArgument valueArgument,
        in int constant,
        in NodeArgument nodeArgument,
        in ConditionArgument conditionArgument,
        in size_t lineNumber,
        in size_t colNumber,
    ) {
        this.instruction = instruction;
        this.valueArgument = valueArgument;
        this.constant = constant;
        this.nodeArgument = nodeArgument;
        this.conditionArgument = conditionArgument;
        this.lineNumber = lineNumber;
        this.colNumber = colNumber;
    }
    
    // Get a string representation of this statement.
    // Useful for debugging.
    string toString() const {
        string result = this.instruction.to!string[1 .. $];
        if(
            this.nodeArgument !is NodeArgument.None &&
            this.nodeArgument !is NodeArgument._self
        ) {
            result ~= " " ~ this.nodeArgument.to!string[1 .. $];
        }
        if(
            this.conditionArgument !is ConditionArgument.None &&
            this.conditionArgument !is ConditionArgument._always
        ) {
            result ~= " " ~ this.conditionArgument.to!string[1 .. $];
        }
        if(this.valueArgument is ValueArgument.Constant) {
            result ~= " " ~ this.constant.to!string;
        }else if(this.valueArgument !is ValueArgument.None) {
            result ~= " " ~ this.valueArgument.to!string[1 .. $];
        }
        return result ~ ";";
    }
}

// Data structure passed around by methods of the Node struct
// Useful for storing an incremented or decremented stack pointer in
// combination with an overflow/underflow flag
struct StackOffset {
    size_t index;
    bool wrapped;
}

// Represent a single node within a Jungle program
struct Node {
    enum StackSize = 256;
    
    // Pointer to node's parent
    Node* parent;
    // Pointer to node's left child
    Node* left;
    // Pointer to node's right child
    Node* right;
    // Pointer to the node which last executed "goto" or "transfer" with
    // this node as the destination
    Node* returnTo;
    // Resume execution in returnTo at this program counter location when
    // executing a "return" or "return_with" instruction
    size_t returnPc;
    
    // List of statements given for this node
    Statement[] statements;
    
    // Location in source file (line number)
    size_t lineNumber;
    // Location in source file (column number)
    size_t colNumber;
    
    // Store stack data
    int[StackSize] stack;
    // Current value of the accumulator
    int acc;
    // Program counter/instruction pointer
    size_t pc;
    // Stack pointer
    size_t sp;
    // Carry flag
    int carry;
    // Multiplication overflow
    int overflow;
    // Stack pointer overflow flag
    int wrapped;
    // Divide by zero flag
    int divz;
    // Error flag/code (refer to ErrorCode enum)
    int error;
    
    // Get a string representation. Potentially useful for debugging.
    string toString(in string indent = "") const {
        string result = "";
        foreach(statement; this.statements) {
            result ~= indent ~ statement.toString() ~ '\n';
        }
        if(this.left) {
            result ~= "left (\n" ~ this.left.toString("  " ~ indent) ~ indent ~ ")\n";
        }
        if(this.right) {
            result ~= "right (\n" ~ this.right.toString("  " ~ indent) ~ indent ~ ")\n";
        }
        return result;
    }
    
    // Get a string respresentation of the node's stack data
    string stackToString() const {
        string result = "(" ~ this.sp.to!string ~ ")";
        if(this.sp) result ~= ": ";
        for(size_t i = 0; i < this.sp; i++) {
            if(i) result ~= ", ";
            result ~= this.stack[i].to!string;
        }
        return result;
    }
    
    // Get pointer to sibling node, i.e. parent's other child
    @property Node* sibling() {
        if(!this.parent) return null;
        return this.parent.left is &this ? this.parent.right : this.parent.left;
    }
    
    // Get pointer to root node
    @property Node* root() {
        if(!this.parent) return &this;
        Node* node = this.parent;
        while(node.parent !is null) node = node.parent;
        return node;
    }
    
    // Get pointer to leftmost node
    // Begin with this node then traverse left children
    // Return the first node without a left child
    @property Node* leftmost() {
        if(!this.left) return &this;
        Node* node = this.left;
        while(node.left !is null) node = node.left;
        return node;
    }
    
    // Get pointer to rightmost node
    // Begin with this node then traverse right children
    // Return the first node without a right child
    @property Node* rightmost() {
        if(!this.right) return &this;
        Node* node = this.right;
        while(node.right !is null) node = node.right;
        return node;
    }
    
    // Get pointer to next node in a forward in-order traversal
    @property Node* next() {
        if(this.right) return this.right.leftmost;
        Node* node = &this;
        while(node && node.parent) {
            if(node is node.parent.left) return node.parent;
            node = node.parent;
        }
        return null;
    }
    
    // Get pointer to next node in a reversed in-order traversal
    @property Node* prev() {
        if(this.left) return this.left.rightmost;
        Node* node = &this;
        while(node && node.parent) {
            if(node is node.parent.right) return node.parent;
            node = node.parent;
        }
        return null;
    }
    
    // Given a node argument, determine which node is being referred to
    Node* getNode(in NodeArgument nodeArgument) {
        with(NodeArgument) final switch(nodeArgument) {
            case None: return null;
            case _self: return &this;
            case _root: return this.root;
            case _parent: return this.parent;
            case _left: return this.left;
            case _right: return this.right;
            case _sibling: return this.sibling;
            case _leftmost: return this.leftmost;
            case _rightmost: return this.rightmost;
            case _next: return this.next;
            case _prev: return this.prev;
            case _origin: return this.returnTo;
        }
    }
    
    StackOffset prevSp() const {
        return this.sp <= 0 ? StackOffset(StackSize - 1, 1) : StackOffset(this.sp - 1);
    }
    
    StackOffset swapSp() const {
        if(this.sp <= 0) return StackOffset(StackSize - 2, 1);
        else if(this.sp == 1) return StackOffset(StackSize - 1, 1);
        else return StackOffset(this.sp - 2);
    }
    
    StackOffset nextSp() const {
        return this.sp + 1 >= StackSize ? StackOffset(0, 1) : StackOffset(this.sp + 1);
    }
    
    // Return true when a given condition was satisfied and false otherwise
    bool checkCondition(in ConditionArgument conditionArgument) const {
        with(ConditionArgument) final switch(conditionArgument) {
            case None: return false;
            case _always: return true;
            case _if_zero: return this.acc == 0;
            case _if_nonzero: return this.acc != 0;
            case _if_positive: return this.acc > 0;
            case _if_not_positive: return this.acc <= 0;
            case _if_negative: return this.acc < 0;
            case _if_not_negative: return this.acc >= 0;
            case _if_carry: return this.carry != 0;
            case _if_not_carry: return this.carry == 0;
            case _if_divz: return this.divz != 0;
            case _if_not_divz: return this.divz == 0;
            case _if_wrapped: return this.wrapped != 0;
            case _if_not_wrapped: return this.wrapped == 0;
            case _if_error: return this.error != 0;
            case _if_no_error: return this.error == 0;
        }
    }
    
    Node* step() {
        if(this.pc >= this.statements.length) {
            return null;
        }
        Node* targetNode = void;
        Statement statement = this.statements[this.pc++];
        if(statement.nodeArgument is NodeArgument._origin) {
            targetNode = this.returnTo;
            if(!targetNode) return null;
        }else {
            targetNode = statement.targetNode;
        }
        int arg = void;
        with(ValueArgument) final switch(statement.valueArgument) {
            case None: arg = 0; break;
            case Constant: arg = statement.constant; break;
            case _top: arg = this.stack[this.prevSp().index]; break;
            case _acc: arg = this.acc; break;
            case _carry: arg = this.carry; break;
            case _overflow: arg = this.overflow; break;
            case _divz: arg = this.divz; break;
            case _wrapped: arg = this.wrapped; break;
            case _error: arg = this.error; break;
        }
        with(Instruction) final switch(statement.instruction) {
            case Invalid:
                assert(false);
            case _exit:
                return null;
            case _void:
                break;
            case _goto:
                if(!this.checkCondition(statement.conditionArgument)) return &this;
                assert(targetNode);
                targetNode.returnTo = &this;
                targetNode.returnPc = this.pc;
                targetNode.pc = 0;
                return targetNode;
            case _transfer:
                if(!this.checkCondition(statement.conditionArgument)) return &this;
                assert(targetNode);
                targetNode.returnTo = &this;
                targetNode.returnPc = this.pc;
                targetNode.acc = arg;
                targetNode.pc = 0;
                return targetNode;
            case _again:
                if(this.checkCondition(statement.conditionArgument)) this.pc = 0;
                return &this;
            case _return:
                if(!this.checkCondition(statement.conditionArgument)) return &this;
                if(this.returnTo) {
                    this.returnTo.pc = this.returnPc;
                }
                return this.returnTo;
            case _return_with:
                if(!this.checkCondition(statement.conditionArgument)) return &this;
                if(this.returnTo) {
                    this.returnTo.pc = this.returnPc;
                    this.returnTo.acc = arg;
                }
                return this.returnTo;
            case _swap:
                assert(targetNode);
                const top = targetNode.prevSp();
                const next = targetNode.swapSp();
                assert(top.index < targetNode.stack.length);
                assert(next.index < targetNode.stack.length);
                assert(top.index != next.index);
                const t = targetNode.stack[top.index];
                targetNode.stack[top.index] = targetNode.stack[next.index];
                targetNode.stack[next.index] = t;
                this.wrapped = cast(int) next.wrapped;
                break;
            case _push:
                assert(targetNode);
                const next = targetNode.nextSp();
                targetNode.stack[targetNode.sp] = arg;
                targetNode.sp = next.index;
                this.wrapped = cast(int) next.wrapped;
                break;
            case _pop:
                assert(targetNode);
                const prev = targetNode.prevSp();
                targetNode.sp = prev.index;
                this.acc = targetNode.stack[prev.index];
                this.wrapped = cast(int) prev.wrapped;
                break;
            case _discard:
                assert(targetNode);
                const prev = targetNode.prevSp();
                targetNode.sp = prev.index;
                this.wrapped = cast(int) prev.wrapped;
                break;
            case _peek:
                assert(targetNode);
                const prev = targetNode.prevSp();
                this.acc = targetNode.stack[prev.index];
                this.wrapped = cast(int) prev.wrapped;
                break;
            case _assign:
                assert(targetNode);
                targetNode.acc = arg;
                break;
            case _inc:
                this.carry = (this.acc == int.max);
                this.acc++;
                break;
            case _dec:
                this.carry = (this.acc == int.min);
                this.acc--;
                break;
            case _shl:
                const uint shift = arg & 0x1F;
                this.overflow = this.acc >> (32 - shift);
                this.acc <<= shift;
                this.carry = (this.acc & 0x80000000 ?
                    this.overflow != 0xFFFFFFFF : this.overflow != 0
                );
                break;
            case _shr:
                const uint shift = arg & 0x1F;
                const int mask = ~(0xFFFFFFFF << shift);
                this.overflow = this.acc & mask;
                this.acc >>>= shift;
                this.carry = (this.overflow != 0);
                break;
            case _sar:
                const uint shift = arg & 0x1F;
                const int mask = ~(0xFFFFFFFF << shift);
                this.overflow = this.acc & mask;
                this.acc >>= shift;
                this.carry = (this.overflow != 0);
                break;
            case _negate:
                this.carry = (this.acc == int.min);
                this.acc = -this.acc;
                break;
            case _abs:
                this.carry = (this.acc == int.min);
                this.acc = (this.acc >= 0 ? this.acc : -this.acc);
                break;
            case _not:
                this.acc = ~this.acc;
                break;
            case _add:
                const long sum = (cast(long) this.acc) + arg;
                this.acc = cast(int) sum;
                this.carry = ((cast(long) this.acc) != sum);
                break;
            case _sub:
                const long diff = (cast(long) this.acc) - arg;
                this.acc = cast(int) diff;
                this.carry = ((cast(long) this.acc) != diff);
                break;
            case _mul:
                long product = (cast(long) this.acc) * arg;
                this.overflow = cast(int) (product >>> 32);
                this.acc = cast(int) product;
                this.carry = ((cast(long) this.acc) != product);
                break;
            case _div:
                if(arg != 0) {
                    this.acc /= arg;
                    this.divz = 0;
                }else {
                    this.acc = 0;
                    this.divz = 1;
                }
                break;
            case _mod:
                if(arg != 0) {
                    this.acc = ((this.acc % arg) + arg) % arg;
                    this.divz = 0;
                }else {
                    this.acc = 0;
                    this.divz = 1;
                }
                break;
            case _rem:
                if(arg != 0) {
                    this.acc %= arg;
                    this.divz = 0;
                }else {
                    this.acc = 0;
                    this.divz = 1;
                }
                break;
            case _and:
                this.acc &= arg;
                break;
            case _or:
                this.acc |= arg;
                break;
            case _xor:
                this.acc ^= arg;
                break;
            case _clear_error:
                this.error = 0;
                break;
            case _write_char:
                const isValid = arg >= 0 && arg < 0x110000;
                if(isValid) write(cast(dchar) arg);
                break;
            case _write_int:
                write(arg.to!string);
                break;
            case _read_char:
                auto chars = FileRange(stdin).byUTF!dchar();
                if(!chars.empty) {
                    const dchar front = chars.front;
                    this.acc = cast(int) front;
                }else {
                    this.acc = 0;
                    this.error = NamedConstant._read_char_error;
                    assert(this.error);
                }
                break;
            case _read_int:
                auto parseResult = parseNumber(readln().strip());
                if(!parseResult.error) {
                    this.acc = parseResult.value;
                }else {
                    this.acc = 0;
                    this.error = NamedConstant._read_int_error;
                }
                break;
        }
        return &this;
    }
}

// Represents a Jungle program
struct Program {
    // List of all nodes in the program tree
    Node*[] nodes;
    // The node whose statements are currently being executed (initially the root)
    Node* currentNode;
    // Store errors encountered while loading a program
    StatusList status;
    
    // Get the root node in the program tree
    @property Node* root() {
        assert(this.nodes.length);
        return this.nodes[0];
    }
    
    // Load a Program object from a given source code string
    static load(in string src) {
        Program program;
        program.parse(src);
        if(!program.status.hasError) {
            program.checkSemantics();
        }
        return program;
    }
    
    // Add an error status specifically for a invalid or missing argument
    // To be invoked during statement parsing, upon encountering an arg issue
    void addArgError(T)(in T loc, in InstructionInfo info, in string message) {
        return this.status.addError(loc, message ~ " Expected: " ~ info.toString());
    }
    
    // Build a node tree from a source code string
    // The code will be tokenized
    void parse(in string src) {
        Tokenizer tokenizer = Tokenizer(src);
        tokenizer.parse();
        this.status.addFrom(tokenizer.status);
        if(!this.status.hasError) {
            this.parse(tokenizer.tokens);
        }
    }
    
    // Build a node tree from tokenized source code
    void parse(in Token[] tokens) {
        size_t i = 0;
        // Initialize a stack containing only the root node
        Node* root = new Node();
        root.lineNumber = 1;
        root.colNumber = 1;
        this.nodes = [root];
        Node*[] stack = [root];
        // Helper for skipping tokens up to the next semicolon (due to bad statement)
        void gotoNextStatement() {
            while(i < tokens.length && tokens[i++].type !is Token.Type.Semicolon) {}
        }
        // Parse all the tokens
        while(i < tokens.length && !this.status.hasFatal) {
            const token = tokens[i];
            // Close paren ')' (end of node definition)
            if(token.type is Token.Type.Close) {
                if(stack.length <= 1) {
                    this.status.addFatal(token, StatusMessage.UnbalancedParens);
                    return;
                }else {
                    stack.length--;
                    i++;
                }
            // Redundant semicolon. Gross, but admissable
            }else if(token.type is Token.Type.Semicolon) {
                this.status.addWarning(token, StatusMessage.ExtraSemicolon);
                i++;
            // Found a number literal (but expected a statement or node)
            }else if(token.type is Token.Type.Number) {
                this.status.addError(token, StatusMessage.UnexpectedNumber);
                gotoNextStatement();
            // Found a string literal (but expected a statement or node)
            }else if(token.type is Token.Type.String) {
                this.status.addError(token, StatusMessage.UnexpectedString);
                gotoNextStatement();
            // Found something else that's not a keyword opening a statement/node
            }else if(token.type !is Token.Type.Keyword) {
                this.status.addError(token, StatusMessage.BadStatement);
                gotoNextStatement();
            // Found a statement with an instruction, as opposed to a child node
            }else if(token.content != "left" && token.content != "right") {
                auto statements = this.parseStatement(tokens, &i);
                stack[$ - 1].statements ~= statements;
                // Empty list implies parsing error; skip to next statement
                if(!statements.length) gotoNextStatement();
            // Found "left" or "right" but not followed by an open paren
            }else if(!stack.length || i + 1 >= tokens.length ||
                tokens[i + 1].type !is Token.Type.Open
            ) {
                this.status.addError(token, StatusMessage.BadNode);
                gotoNextStatement();
            // Found "left" or "right" i.e. a child node declaration
            }else {
                assert(stack.length);
                assert(stack[0] is this.nodes[0]);
                Node* newNode = new Node(stack[$ - 1]);
                this.nodes ~= newNode;
                newNode.lineNumber = token.lineNumber;
                newNode.colNumber = token.colNumber;
                // Left-hand child node declaration
                if(token.content == "left") {
                    if(stack[$ - 1].left !is null) {
                        this.status.addError(token, StatusMessage.RepeatLeft);
                    }else {
                        (*stack[$ - 1]).left = newNode;
                    }
                // Right-hand child node declaration
                }else if(token.content == "right") {
                    if(stack[$ - 1].right !is null) {
                        this.status.addError(token, StatusMessage.RepeatRight);
                    }else {
                        (*stack[$ - 1]).right = newNode;
                    }
                // This shouldn't happen
                }else {
                    this.status.addFatal(token, StatusMessage.BadNode);
                    return;
                }
                stack ~= newNode;
                i += 2;
            }
        }
        // Make sure everything is working correctly
        if(stack.length != 1 && !this.status.hasFatal) {
            this.status.addFatal(tokens[0], StatusMessage.UnbalancedParens);
        }
    }
    
    // Parse a single statement in source code
    // Note that statements with more than one value argument are broken
    // into multiple Statement objects, one for each argument
    Statement[] parseStatement(in Token[] tokens, size_t* i) {
        if(!tokens.length) return [];
        // Remember initial value of i
        size_t initial = *i;
        // Represent a single value argument
        struct Value {ValueArgument type; int constant;}
        // Parse first token and get instruction info, e.g. expected parameters
        Instruction instruction = getInstruction(tokens[*i].content);
        InstructionInfo info = InstructionInfo.get(instruction);
        // Check for invalid/unrecognized instruction
        if(instruction is Instruction.Invalid) {
            this.status.addError(tokens[initial], StatusMessage.BadInstruction);
            return [];
        }
        // Declare vars to store arguments
        Value[] values;
        NodeArgument nodeArgument = NodeArgument.None;
        ConditionArgument conditionArgument = ConditionArgument.None;
        // Parse the arguments
        (*i)++;
        while(*i < tokens.length && !this.status.hasFatal) {
            const arg = tokens[*i];
            // Reached a semicolon (end of statement)
            if(arg.type is Token.Type.Semicolon) {
                (*i)++;
                break;
            // Found a number literal constant
            }else if(arg.type is Token.Type.Number && info.needsValue) {
                values ~= Value(ValueArgument.Constant, arg.value);
            // Found a string literal (treated as a list of constants)
            }else if(arg.type is Token.Type.String && info.needsValue) {
                foreach(ch; arg.content.byUTF!dchar) {
                    values ~= Value(ValueArgument.Constant, cast(int) ch);
                }
            // Unexpected close paren (unterminated statement)
            }else if(arg.type is Token.Type.Close) {
                this.status.addError(arg, StatusMessage.UnterminatedStatement);
                break;
            // Unexpected open paren
            }else if(arg.type is Token.Type.Open) {
                this.status.addError(arg, StatusMessage.UnexpectedParen);
                break;
            // Catch-all for any other unexpected token type (shouldn't happen though)
            }else if(arg.type !is Token.Type.Keyword) {
                this.addArgError(arg, info, StatusMessage.BadArgument);
                return [];
            // Keyword is a named constant e.g. "min" and "max"?
            }else if(info.needsValue && getNamedConstant(arg.content)) {
                values ~= Value(ValueArgument.Constant, getNamedConstant(arg.content));
            // Keyword is the "no_error" named constant?
            }else if(info.needsValue && arg.content == "no_error") {
                values ~= Value(ValueArgument.Constant, 0);
            // Keyword is a special value argument e.g. "top", "acc"?
            }else if(info.needsValue && getValueArgument(arg.content)) {
                values ~= Value(getValueArgument(arg.content));
            // Keyword is a node argument e.g. "self", "parent"?
            }else if(info.nodeArgument &&
                nodeArgument is NodeArgument.None &&
                getNodeArgument(arg.content)
            ) {
                nodeArgument = getNodeArgument(arg.content);
                assert(nodeArgument !is NodeArgument.None);
            // Keyword is a condition argument e.g. "if_zero"?
            }else if(info.conditionArgument &&
                conditionArgument is ConditionArgument.None &&
                getConditionArgument(arg.content)
            ) {
                conditionArgument = getConditionArgument(arg.content);
                assert(conditionArgument !is ConditionArgument.None);
            // Keyword is an instruction (probably an accidental missing semicolon)
            }else if(getInstruction(arg.content)) {
                this.status.addError(arg, StatusMessage.UnexpectedInstruction);
            // Unrecognized keyword
            }else {
                this.status.addError(arg, StatusMessage.BadKeywordArgument);
                return [];
            }
            // Too many constants for this instruction?
            if(values.length > 1 && !info.multiValue) {
                this.addArgError(arg, info, StatusMessage.OnlyOneArgument);
                return [];
            }
            (*i)++;
        }
        // Ensure no missing arguments
        if(info.needsValue && values.length <= 0) {
            this.addArgError(tokens[initial], info, StatusMessage.MissingArgument);
            return [];
        }
        // Make sure the statement was terminated by a semicolon (not e.g. EOF)
        if(*i == tokens.length && tokens[(*i) - 1].type !is Token.Type.Semicolon) {
            this.status.addError(tokens[initial], StatusMessage.UnterminatedStatement);
        }
        // All is well; go ahead and build the statement or statements
        if(!nodeArgument) nodeArgument = NodeArgument._self;
        if(!conditionArgument) conditionArgument = ConditionArgument._always;
        if(values.length) {
            assert(info.multiValue || values.length == 1);
            Statement[] statements;
            statements.reserve(values.length);
            for(size_t j = 0; j < values.length; j++) {
                const value = (info.valueCount is ValueCount.MultiReverse ?
                    values[values.length - j - 1] : values[j]
                );  
                statements ~= Statement(
                    instruction, value.type, value.constant,
                    nodeArgument, conditionArgument,
                    tokens[initial].lineNumber, tokens[initial].colNumber
                );
            }
            return statements;
        }else {
            return [Statement(
                instruction, ValueArgument.None, 0,
                nodeArgument, conditionArgument,
                tokens[initial].lineNumber, tokens[initial].colNumber
            )];
        }
    }
    
    // Run after parsing step to check for valid semantics
    // Accumulates problems in the Program.errors list
    // Also fills in Statement.targetNode pointers.
    // This has to be done before the program will run!
    void checkSemantics() {
        // Bail out immediately if the program didn't load correctly
        if(this.status.hasError) return;
        // Check semantics on each program node
        foreach(node; this.nodes) this.checkNodeSemantics(node);
    }
    
    // Called by checkSemantics for each node in the program
    void checkNodeSemantics(Node* node) {
        // Warning for nodes with no statements or children
        // Workaround for warning: Use "void" e.g. "left (void;)"
        if(!node.statements.length && node.left is null && node.right is null) {
            const message = (node is this.root ?
                StatusMessage.EmptyRootNode : StatusMessage.EmptyNode
            );
            this.status.addWarning(node, message);
        }
        // Check every statement
        foreach(ref statement; node.statements) {
            // Check validity of target node arguments
            if(statement.nodeArgument !is NodeArgument._origin) {
                auto target = node.getNode(statement.nodeArgument);
                statement.targetNode = target;
                if(!target) {
                    this.status.addError(statement, StatusMessage.BadNodeArgument);
                }
            }
        }
    }
    
    // Fully execute the program
    void execute() {
        // Don't do anything if there was an error loading the program
        if(this.status.hasError) return;
        // Begin execution at the root node
        if(this.nodes.length) this.currentNode = this.nodes[0];
        // Keep executing instructions until one of them terminates the program
        while(!this.step()) {}
    }
    
    // Execute the next statement in the program
    // Returns true when the statement caused the program to terminate
    // Returns false when execution should continue to the next statement
    bool step() {
        if(!this.nodes.length || !this.currentNode) {
            return true;
        }
        Node* nextNode = this.currentNode.step();
        if(nextNode is null) {
            return true;
        }else if(nextNode !is this.currentNode) {
            this.currentNode = nextNode;
        }
        return false;
    }
}
