import core.thread;
import std.array;
import std.conv;
import std.datetime;
import std.file;
import std.process;
import std.stdio;
import std.string;

// Represent test cases loaded from a single file in the tests/ directory
struct Test {
    string path;
    string[] inputs;
    string[] outputs;
    int[] expectStatus;
    string source;
    bool passed;
}

// Look for the Jungle interpreter binary
string getBinary() {
    version(Windows) {
        enum tryPaths = [
            "bin/jungle.exe",
            "bin/jungle",
            "jungle.exe",
            "jungle",
        ];
    }else {
        enum tryPaths = [
            "./bin/jungle",
            "./jungle",
            "jungle",
        ];
    }
    foreach(path; tryPaths) {
        if(path.exists) return path;
    }
    return "";
}

// Do the things
int main(in string[] args) {
    // Find binary (or accept path as first argument)
    string testsPath = args.length > 1 ? args[1] : "tests";
    string binaryPath = args.length > 2 ? args[2] : getBinary();
    if(!binaryPath || !binaryPath.exists) {
        writeln("Unable to find Jungle binary.");
        writeln("Try passing the binary path as an argument.");
        writeln("Example: run_tests path/to/tests path/to/jungle");
        return 1;
    }
    // Find all *.ck files in the tests/ directory
    Test[] tests;
    const timestamp = Clock.currTime().toISOString();
    auto entries = dirEntries(testsPath, "*.jg", SpanMode.shallow);
    foreach(filePath; entries) {
        Test test = Test(filePath);
        uint state = 0;
        foreach(line; File(filePath).byLine) {
            if(line == "///INPUT") { // Provided input
                test.inputs ~= ""; state = 1;
            }else if(line == "///OUTPUT") { // Expected output (exit code = 0)
                test.outputs ~= ""; state = 2;
                test.expectStatus ~= 0;
            }else if(line == "///ERROR") { // Expected output (exit code = 1)
                test.outputs ~= ""; state = 2;
                test.expectStatus ~= 1;
            }else if(line == "///SOURCE") { // Jungle source
                state = 3;
            }else if(state == 1) {
                test.inputs[$ - 1] ~= line ~ "\n";
            }else if(state == 2) {
                test.outputs[$ - 1] ~= line ~ "\n";
            }else if(state == 3) {
                test.source ~= line ~ "\n";
            }
        }
        if(test.outputs.length && test.source.length) {
            tests ~= test;
        }
    }
    // Now run the tests
    writeln("Running ", tests.length, " tests:");
    stdout.flush();
    uint passedCount = 0;
    // For each test...
    foreach(test; tests) {
        test.passed = true;
        // For each expected output (e.g. given different inputs)...
        for(size_t i = 0; i < test.outputs.length; i++) {
            // Write temporary input and output files
            const istr = to!string(1 + i);
            string inputPath = test.path ~ "." ~ istr ~ "." ~ timestamp ~ ".in.log";
            string outputPath = test.path ~ "." ~ istr ~ "." ~ timestamp ~ ".out.log";
            auto inputFile = File(inputPath, "wb");
            auto outputFile = File(outputPath, "wb");
            inputFile.write(i < test.inputs.length ? test.inputs[i] : "");
            inputFile.open(inputPath, "rb");
            // Run Jungle interpreter with given input, output, and source
            auto testProcess = spawnProcess(
                [binaryPath, "eval", test.source], inputFile, outputFile
            );
            // Wait for completion (5s timeout)
            enum TotalTimeout = 5000;
            enum WaitIntervals = 50;
            enum IntervalSleep = TotalTimeout / WaitIntervals;
            int testStatus = 0;
            bool testCompleted = false;
            for(size_t j = 0; j < WaitIntervals && !testCompleted; j++) {
                Thread.sleep(dur!"msecs"(IntervalSleep));
                auto result = tryWait(testProcess);
                if(result.terminated) {
                    testStatus = result.status;
                    testCompleted = true;
                    break;
                }
            }
            if(!testCompleted) {
                kill(testProcess);
                wait(testProcess);
            }
            // Cleanup
            inputFile.close();
            inputPath.remove();
            outputFile.close();
            auto output = cast(string) outputPath.read();
            const expectedNorm = test.outputs[i].replace("\r\n", "\n").strip();
            const actualNorm = output.replace("\r\n", "\n").strip();
            // Report failures and keep the output log on test fail, for debugging
            if(!testCompleted) {
                writeln("TEST FAILURE: ", test.path, " (Timed out) (input #", i + 1, ")");
                test.passed = false;
            }else if(testStatus && !test.expectStatus[i]) {
                writeln("TEST FAILURE: ", test.path, " (Nonzero status) (input #", i + 1, ")");
                test.passed = false;
            }else if(expectedNorm != actualNorm) {
                writeln("TEST FAILURE: ", test.path, " (Wrong output) (input #", i + 1, ")");
                test.passed = false;
            }else if(testStatus != test.expectStatus[i]) {
                writeln("TEST FAILURE: ", test.path, " (Wrong exit status) (input #", i + 1, ")");
                test.passed = false;
            }else {
                writeln("Test passed: ", test.path, " (input #", istr, ")");
                outputPath.remove(); // Remove output log on successful test
            }
            stdout.flush();
        }
        passedCount += test.passed;
    }
    // All done!
    writeln(passedCount, " of ", tests.length, " tests passed.");
    return passedCount == tests.length ? 0 : 1;
}
