import std.conv;

// Common status message strings
struct StatusMessage {
    enum string BadStatement = "Malformed statement.";
    enum string BadInstruction = "Unrecognized instruction.";
    enum string UnexpectedInstruction = "Encountered an instruction but expected an argument.";
    enum string BadKeywordArgument = "Encountered an unrecognized keyword.";
    enum string BadArgument = "Illegal argument.";
    enum string MissingArgument = "Statement is missing a required value argument.";
    enum string OnlyOneArgument = "Instruction cannot accept more than one value argument.";
    enum string UnexpectedNumber = "Unexpected number literal.";
    enum string UnexpectedString = "Unexpected string literal.";
    enum string UnbalancedParens = "Unbalanced parentheses.";
    enum string UnexpectedParen = "Unexpected parenthese.";
    enum string UnterminatedString = "Unterminated string literal.";
    enum string BadIntLiteral = "Malformed integer literal.";
    enum string BadHexLiteral = "Malformed hexadecimal literal.";
    enum string BadEscape = "Malformed escape sequence in string literal.";
    enum string BadNode = "Malformed child node declaration.";
    enum string RepeatLeft = "Repeated left node declaration.";
    enum string RepeatRight = "Repeated right node declaration.";
    enum string BadNodeArgument = "Node argument refers to nonexistent node.";
    enum string ExtraSemicolon = "Redundant semicolon.";
    enum string EmptyNode = "Node contains no statements.";
    enum string EmptyRootNode = "Root program node contains no statements.";
    enum string UnterminatedStatement = "Statement is missing a terminating semicolon.";
}

// Enumerate Status severity levels
enum Severity: uint {
    Info = 0,
    Warning = 1,
    Error = 2,
    Fatal = 3,
}

// Represents a status message encountered while loading a Jungle program
struct Status {
    Severity severity;
    size_t lineNumber;
    size_t colNumber;
    string message;
    
    this(T)(in Severity severity, in T loc, in string message) {
        this(severity, loc.lineNumber, loc.colNumber, message);
    }
    
    this(
        in Severity severity,
        in size_t lineNumber,
        in size_t colNumber,
        in string message
    ) {
        this.severity = severity;
        this.lineNumber = lineNumber;
        this.colNumber = colNumber;
        this.message = message;
    }
    
    // Get a nice human-readable representation of this status. Looks like:
    // Severity: Pertinent status message (L1:1)
    string toString() const {
        return this.severity.to!string ~ ": " ~ this.message ~ " (L" ~ (
            this.lineNumber.to!string ~ ":" ~ this.colNumber.to!string
        ) ~ ")";
    }
}

struct StatusList {
    Status[][4] items;
    
    void add(Status status) {
        // Don't add exact duplicates of existing status messages
        if(this.has(status)) return;
        // Append to the appropriate severity list
        this.items[status.severity] ~= status;
    }
    
    void add(T)(in Severity severity, in T loc, in string message) {
        return this.add(Status(severity, loc, message));
    }
    
    void add(
        in Severity severity,
        in size_t lineNumber, in size_t colNumber, in string message
    ) {
        return this.add(Status(severity, lineNumber, colNumber, message));
    }
    
    void addInfo(in size_t lineNumber, in size_t colNumber, in string message) {
        return this.add(Severity.Info, lineNumber, colNumber, message);
    }
    
    void addInfo(T)(in T loc, in string message) {
        return this.add(Severity.Info, loc, message);
    }
    
    void addWarning(in size_t lineNumber, in size_t colNumber, in string message) {
        return this.add(Severity.Warning, lineNumber, colNumber, message);
    }
    
    void addWarning(T)(in T loc, in string message) {
        return this.add(Severity.Warning, loc, message);
    }
    
    void addError(in size_t lineNumber, in size_t colNumber, in string message) {
        return this.add(Severity.Error, lineNumber, colNumber, message);
    }
    
    void addError(T)(in T loc, in string message) {
        return this.add(Severity.Error, loc, message);
    }
    
    void addFatal(in size_t lineNumber, in size_t colNumber, in string message) {
        return this.add(Severity.Fatal, lineNumber, colNumber, message);
    }
    
    void addFatal(T)(in T loc, in string message) {
        return this.add(Severity.Fatal, loc, message);
    }
    
    void addFrom(in StatusList list) {
        for(size_t i = 0; i <= Severity.Fatal; i++) {
            foreach(status; list.items[i]) this.items[i] ~= status;
        }
    }
    
    // True when there is any status of the given severity or higher
    bool has(in Severity severity) const {
        for(size_t i = cast(size_t) severity; i <= Severity.Fatal; i++) {
            if(this.items[i].length > 0) return true;
        }
        return false;
    }
    
    // True when the StatusList already contains an exactly matching status
    bool has(in Status status) const {
        foreach(ownStatus; this.items[cast(size_t) status.severity]) {
            if(ownStatus == status) return true;
        }
        return false;
    }
    
    // True when the program encountered at least one error status
    bool hasError() const {
        return this.items[cast(size_t) Severity.Error].length > 0;
    }
    
    // True when the program encountered at least one fatal status
    bool hasFatal() const {
        return this.items[cast(size_t) Severity.Fatal].length > 0;
    }
    
    // Get a string showing every error, separated by newlines
    string toString() const {
        string result = "";
        for(size_t i = 1 + cast(size_t) Severity.Fatal; i > 0; i--) {
            foreach(status; this.items[i - 1]) result ~= status.toString() ~ "\n";
        }
        return result;
    }
}
