import std.conv;
import std.string;

import contains : contains;
import parsenumber : parseHex, parseNumber, NumberFormat;
import status : Severity, Status, StatusList, StatusMessage;

// Represents a single token in source code
struct Token {
    enum Type {
        None, // No token/invalid token
        Keyword, // Keyword e.g. "goto_right"
        Number, // Number literal e.g. "1234" or "0xFFFF"
        String, // Double quoted string literal
        Semicolon, // Semicolon ';'
        Open, // Open paren '('
        Close, // Close paren ')'
    }
    
    Type type;
    size_t lineNumber;
    size_t colNumber;
    string content;
    int value;
    
    string toString() const {
        return this.dataToString ~ " @ " ~ this.locationToString();
    }
    
    string dataToString() const {
        with(Type) final switch(this.type) {
            case None: return "none";
            case Keyword: return "keyword: " ~ this.content;
            case Number: return "number: " ~ this.value.to!string;
            case String: return "string: " ~ this.content;
            case Semicolon: return "semicolon";
            case Open: return "open_paren";
            case Close: return "close_paren";
        }
    }
    
    string locationToString() const {
        return this.lineNumber.to!string ~ ":" ~ this.colNumber.to!string;
    }
}

// Utility for tokenizing an input source code string
struct Tokenizer {
    string src;
    StatusList status;
    Token[] tokens;
    Token[] parenStack;
    bool finishedToken = true;
    size_t index = 0;
    size_t tokenStart = 0;
    bool inComment = false;
    bool inQuote = false;
    bool inEscape = false;
    size_t lineNumber = 1;
    size_t lineStart = 0;
    
    // Parse an integer literal
    int parseNumber(T)(in string src, in T loc) {
        auto result = .parseNumber(src);
        if(result.error) {
            with(NumberFormat) final switch(result.format) {
                case Decimal: this.status.addError(loc, StatusMessage.BadIntLiteral); break;
                case Hexadecimal: this.status.addError(loc, StatusMessage.BadHexLiteral); break;
            }
            return 0;
        }else {
            return result.value;
        }
    }
    
    // Convert escape sequences in a string to the characters they represent
    string unescapeString(T)(in string src, in T loc) {
        string content = "";
        size_t i = 1;
        while(i < src.length && !this.status.hasFatal) {
            if(src[i] == '\\') {
                const result = this.unescapeChar(src[i + 1 .. $], loc);
                content ~= result.ch;
                i += 1 + result.length;
            }else {
                content ~= src[i];
                i++;
            }
        }
        assert(i == src.length);
        return content;
    }
    
    // Convert an escape sequence to a single character
    // Receives characters following the '\'
    auto unescapeChar(T)(in string src, in T loc) {
        struct Result {
            char ch = 0;
            size_t length = 1;
        }
        if(!src.length) {
            this.status.addError(loc, StatusMessage.BadEscape);
            return Result('\0', 0);
        }
        switch(src[0]) {
            case '0': return Result(0);
            case 'a': return Result('\a');
            case 'b': return Result('\b');
            case 'e': return Result(27);
            case 'f': return Result('\f');
            case 'n': return Result('\n');
            case 'r': return Result('\r');
            case 't': return Result('\t');
            case 'v': return Result('\v');
            case 'x':
                if(src.length < 4) {
                    this.status.addError(loc, StatusMessage.BadEscape);
                    return Result('\0', src.length);
                }else {
                    auto result = parseHex(src[2 .. 4]);
                    if(result.error) this.status.addError(loc, StatusMessage.BadEscape);
                    return Result(cast(char) result.value, 4);
                }
            default:
                return Result(src[0]);
        }
    }
    
    void addToken(in Token.Type type) {
        const size_t col = 1 + this.index - this.lineStart;
        this.tokens ~= Token(type, this.lineNumber, col);
        this.finishedToken = true;
    }
    
    void startToken(in Token.Type type) {
        const size_t col = 1 + this.index - this.lineStart;
        this.tokens ~= Token(type, this.lineNumber, col);
        this.tokenStart = this.index;
        this.finishedToken = false;
    }
    
    void endToken(in size_t offset) {
        if(this.finishedToken || !this.tokens.length) return;
        Token* token = &this.tokens[$ - 1];
        const string content = this.src[this.tokenStart .. offset + this.index];
        this.finishedToken = true;
        with(Token.Type) final switch(token.type) {
            case None:
                assert(false);
            case Keyword:
                token.content = content;
                break;
            case Number:
                token.value = Tokenizer.parseNumber(content, token);
                break;
            case String:
                token.content = Tokenizer.unescapeString(content[0 .. $ - 1], token);
                break;
            case Semicolon:
                break; // Do nothing
            case Open:
                this.parenStack ~= *token;
                break;
            case Close:
                if(this.parenStack.length) {
                    this.parenStack.length--;
                }else {
                    this.status.addFatal(*token, StatusMessage.UnbalancedParens);
                }
                break;
        }
    }
    
    @property Token currentToken() const {
        return (!this.finishedToken && this.tokens.length ?
            this.tokens[$ - 1] : Token(Token.Type.None)
        );
    }
    
    @property Token.Type currentType() const {
        return (!this.finishedToken && this.tokens.length ?
            this.tokens[$ - 1].type : Token.Type.None
        );
    }
    
    void parse() {
        const begin = this.src.indexOf("///BEGIN///");
        if(begin >= 0) this.index = begin + 11;
        const end = this.src[this.index .. $].indexOf("///END///");
        if(end >= 0) this.src = this.src[0 .. end];
        while(this.index < this.src.length && !this.status.hasFatal) {
            const ch = this.src[this.index];
            const next = (this.index + 1 < this.src.length ?
                this.src[this.index + 1] : '\0'
            );
            if(ch == '\n') {
                this.lineNumber++;
                this.lineStart = 1 + this.index;
            }
            if(this.inEscape) {
                this.inEscape = false;
            }else if(this.inQuote) {
                if(ch == '"') {
                    this.inQuote = false;
                    this.endToken(1);
                }else if(ch == '\\') {
                    this.inEscape = true;
                }
            }else if(this.inComment) {
                this.inComment = (ch != '\n');
            }else if(ch == '/' && next == '/') {
                this.endToken(0);
                this.inComment = true;
            }else if(ch == '"') {
                this.inQuote = true;
                this.endToken(0);
                this.startToken(Token.Type.String);
            }else if(ch == ';') {
                this.endToken(0);
                this.addToken(Token.Type.Semicolon);
            }else if(ch == '(') {
                this.endToken(0);
                this.addToken(Token.Type.Open);
            }else if(ch == ')') {
                this.endToken(0);
                this.addToken(Token.Type.Close);
            }else if("+-0123456789".contains(ch)) {
                if(this.currentType !is Token.Type.Number) {
                    this.endToken(0);
                    this.startToken(Token.Type.Number);
                }
            }else if(" \t\r\n".contains(ch)) {
                this.endToken(0);
            }else if(this.currentType is Token.Type.None) {
                this.startToken(Token.Type.Keyword);
            }
            this.index++;
        }
        if(this.currentType is Token.Type.String) {
            const token = this.currentToken();
            this.status.addFatal(token, StatusMessage.UnterminatedString);
        }else if(this.parenStack.length) {
            const token = this.parenStack[$ - 1];
            this.status.addFatal(token, StatusMessage.UnbalancedParens);
        }else {
            this.endToken(0);
        }
    }
}
